from typing import List

import matplotlib.pyplot as plt
import numpy as np


def euclidean(vector1: List, vector2: List) -> float:
    """
    This method calculates the euclidean distance.

    :param vector1: Vector as list
    :param vector2: Vector as list
    :return: Euclidean distance between vector1 and vector2.
    """
    return np.linalg.norm(np.subtract(vector1, vector2))


def pairwise_arg_min(X: List, Y: List) -> np.ndarray:
    """
    This method returns a list of all pairwise distances from X to Y.

    :param X: Vector with features
    :param Y: Centroids
    :return: List of all pairwise distances from X to Y.
    """
    return np.asarray([np.argmin([euclidean(x, y) for y in Y]) for x in X])


def find_clusters_with_fix_init_centers(X, n_clusters, centers):
    """
    This method finds all clusters.

    :param centers: pre defined centers
    :param X: Data to be clustered
    :param n_clusters: amount of clusters
    :return: All labels for clustering and the centroids.
    """
    while True:
        labels = pairwise_arg_min(X, centers)
        new_centers = np.array([X[labels == i].mean(0) for i in range(n_clusters)])

        if np.all(centers == new_centers):
            break
        centers = new_centers

    return centers, labels


if __name__ == '__main__':
    X = np.array([[1, 1],
                  [1, 4],
                  [2, 2],
                  [10, 3],
                  [11, 2],
                  [11, 4],
                  [4, 12],
                  [6, 11],
                  [7, 10],
                  [8, 10]])

    centers, labels = find_clusters_with_fix_init_centers(X, 3, [X[2], X[3], X[7]])
    plt.scatter(X[:, 0], X[:, 1], c=labels, cmap='viridis')
    plt.scatter(centers[:, 0], centers[:, 1], c='Red', marker='X')
    plt.savefig('./images/a.jpg', dpi=300, bbox_inches='tight')
    plt.close()

    centers, labels = find_clusters_with_fix_init_centers(X, 3, [X[6], X[7], X[8]])
    plt.scatter(X[:, 0], X[:, 1], c=labels, cmap='viridis')
    plt.scatter(centers[:, 0], centers[:, 1], c='Red', marker='X')
    plt.savefig('./images/b.jpg', dpi=300, bbox_inches='tight')
    plt.close()

    X = X[::-1]
    centers, labels = find_clusters_with_fix_init_centers(X, 3, [X[2], X[3], X[7]])
    plt.scatter(X[:, 0], X[:, 1], c=labels, cmap='viridis')
    plt.scatter(centers[:, 0], centers[:, 1], c='Red', marker='X')
    plt.savefig('./images/c.jpg', dpi=300, bbox_inches='tight')
    plt.close()
