from tkinter import Label, Tk, Button, Entry, IntVar
from tkinter.filedialog import askopenfilename
from tokenize import String

from PIL import ImageTk, Image

from orb import ORB


class GUI(object):

    def __init__(self, master):
        """
        This class defines the GUI as one.
        :param master: The Tk root.
        """
        self.master = master
        self.master.title('Filter')

        self.image_label_width = 500
        self.image_label_height = 500

        # Images
        self.image_path = './images/original.jpeg'
        self.edit_path = './images/result.jpeg'
        self.image = self.read_image(self.image_path)

        # Labels
        self.original_image_label = Label(self.master, image=self.image)
        self.original_image_label.pack()
        self.filtered_image_label = Label(self.master, image=self.image)
        self.filtered_image_label.pack()

        # Buttons
        self.file_dialog = Button(master, text='Choose', command=lambda: self.set_new_file())
        self.doit = Button(master, text='Do it', command=lambda: self.change_filter())

        # Entry
        self.k_value = IntVar(self.master)
        self.k_value.set(1)
        self.b_menu = Entry(self.master, textvariable=self.k_value)

        # Layout
        self.file_dialog.grid(row=0, column=0)
        self.doit.grid(row=0, column=1)
        self.b_menu.grid(row=0, column=2)
        self.original_image_label.grid(row=1, column=0)
        self.filtered_image_label.grid(row=1, column=2)

    def read_image(self, path: String) -> ImageTk.PhotoImage:
        """
        This reads an image from a given path and resize it.

        :param path: The path to the image.
        :return: A image to be used in the frontend
        """
        return ImageTk.PhotoImage(
            Image.open(path).resize(
                (self.image_label_width, self.image_label_height),
                Image.ANTIALIAS
            )
        )

    def _update_label_image(self, label: Label, new_image, row: int, column: int) -> None:
        """
        This method updates to label with a new image.

        :param label: The reference to the label which should be changes.
        :param new_image: The reference to the new image.
        :param row: Row in the GUI.
        :param column: Column in the GUI.
        :return: None
        """
        label.destroy()
        label = Label(self.master, image=new_image)
        label.image = new_image
        label.pack()
        label.grid(row=row, column=column)

    def set_new_file(self, *args) -> None:
        """
        This sets the new file and refreshed the paths as well with the lables.

        :return: None
        """

        self.image_path: String = askopenfilename(initialdir='/', title='Select file')

        new_image = self.read_image(self.image_path)

        self._update_label_image(self.original_image_label, new_image, 1, 0)

        self.k_value.set(1)
        self._update_label_image(self.filtered_image_label, new_image, 1, 2)

    def change_filter(self, *args) -> None:
        """
        This changes the filter and updates the label for the edited images.

        :param args: Arguments of the listener.
        :return: None
        """
        self.edit_image()
        edited_image = self.read_image(self.edit_path)
        self._update_label_image(self.filtered_image_label, edited_image, 1, 2)

    def edit_image(self) -> None:
        """
        This method edits a given image and saves the edited one.

        :return: None
        """
        print('From: ' + self.image_path +
              ' to ' + self.edit_path +
              ' and K: '
              + str(self.k_value.get()))

        ORB(
            self.image_path, self.edit_path, self.k_value.get()
        ).get_keypoints()


if __name__ == '__main__':
    """
    This program is for editing images with ORB keypoints.
    Those keypoints are then been grouped with k-means.
    
    Preparation:
        Make sure you have already installed all packages by running:
        $ pip3 install -r requirements.txt
    
    To use the program run:
        $ python3 GUI.py
    
    Notice: The edited image will be saved in ./images/result.py. Each new iteration will overwrite this image. This may
    take a while.
    """
    root = Tk()
    root.resizable(False, False)
    my_gui = GUI(root)
    root.mainloop()
