import os
from tokenize import String
from typing import List, Dict

import cv2 as cv
import matplotlib.pyplot as plt
import numpy as np


class ORB(object):

    def __init__(self, origin_path: String, edit_path: String, k: int = 0):
        """
        This class is the editor which edits a given image and stores a new one.

        :param origin_path: The original image which should be edited.
        :param edit_path: The edited image which should be stored.
        :param k: Is used fr K-Means.
        """
        self.image_path = origin_path
        self.edit_path = edit_path
        self.k = k

    def load_images(self, dir: String) -> List:
        """
        This method concats all images in a dir

        :param dir: The directory to search with all images
        :return: list of all images in a directory
        """
        images = []
        files = os.listdir(dir)
        for file in files:
            img = os.path.join(dir, file)
            img = cv.imread(img, 0)
            if img is not None:
                images.append(img)
        return images

    @staticmethod
    def euclidean(vector1: List, vector2: List) -> float:
        """
        This method calculates the euclidean distance.

        :param vector1: Vector as list
        :param vector2: Vector as list
        :return: Euclidean distance between vector1 and vector2.
        """
        return np.linalg.norm(np.subtract(vector1, vector2))

    def pairwise_arg_min(self, X: List, Y: List) -> np.ndarray:
        """
        This method returns a list of all pairwise distances from X to Y.

        :param X: Vector with features
        :param Y: Centroids
        :return: List of all pairwise distances from X to Y.
        """
        return np.asarray([np.argmin([self.euclidean(x, y) for y in Y]) for x in X])

    def find_clusters(self, X, n_clusters, rseed=2):
        """
        This method finds all clusters.

        :param X: Data to be clustered
        :param n_clusters: amount of clusters
        :param rseed: random seed
        :return: All labels for clustering and the centroids.
        """
        rng = np.random.RandomState(rseed)
        i = rng.permutation(X.shape[0])[:n_clusters]
        centers = X[i]
        while True:
            labels = self.pairwise_arg_min(X, centers)
            new_centers = np.array([X[labels == i].mean(0) for i in range(n_clusters)])

            if np.all(centers == new_centers):
                break
            centers = new_centers

        return centers, labels

    def similar_features(self, descriptors: List, centroids: List, labels: np.ndarray) -> List:
        """
        This calculates the similar features between the groundtruth and a given image.

        :param descriptors: Descriptors of the image
        :param centroids: Centroids of the groundtruth
        :param labels: Labels of the clusters in the groundtruth
        :return: List of similar features by count
        """
        labels = list(set(labels))
        zeros = [0] * len(labels)
        cluster_counts = dict(zip(labels, zeros))

        for des in descriptors:
            min_index = self.pairwise_arg_min([des], centroids)[0]
            cluster_counts[min_index] = cluster_counts[min_index] + 1

        return self.dict_to_list(cluster_counts)

    @staticmethod
    def dict_to_list(target: Dict) -> List:
        """
        Turns a dict in a list.

        :param target: Dict to be turned into a list.
        :return: A list containing all values of a dict
        """
        return [value for _, value in target.items()]

    def get_keypoints(self) -> None:
        """
        This method does K-Means with the ORB Keypoints.
        :return: None
        """

        images = self.load_images("./images/astronauts")
        orb = cv.ORB_create()
        descriptors = None
        for img in images:
            kp, des = orb.detectAndCompute(img, None)
            if descriptors is None:
                descriptors = des
            else:
                descriptors = np.concatenate((descriptors, des), axis=0)

        img = cv.imread(self.image_path, 0)

        kp, des = orb.detectAndCompute(img, None)
        centroids, labels = self.find_clusters(descriptors, self.k)
        hist_with_bins = self.similar_features(des, centroids, labels)

        plt.bar(range(len(hist_with_bins)), np.multiply(1 / sum(hist_with_bins), hist_with_bins))
        plt.savefig(self.edit_path, dpi=300, bbox_inches='tight')
        plt.close()
